T_COST_BYTE_SIZE = 4
M_COST_BYTE_SIZE = 4
P_COST_BYTE_SIZE = 1
N_BITS_BYTE_SIZE = 1
HASH_LEN_CHARS_BYTE_SIZE = 1
COMPRESSION_LEVEL_BYTE_SIZE = 1
COMPRESSION_TYPE_BYTE_SIZE = 1
EXPRESS_POW_BYTE_SIZE = 1

# Chunk slicing to calculate only once per import
M_COST_CHUNK_SLICE = T_COST_BYTE_SIZE + M_COST_BYTE_SIZE
P_COST_CHUNK_SLICE = M_COST_CHUNK_SLICE + P_COST_BYTE_SIZE
N_BITS_CHUNK_SLICE = P_COST_CHUNK_SLICE + N_BITS_BYTE_SIZE
HASH_LEN_CHARS_CHUNK_SLICE = N_BITS_CHUNK_SLICE + HASH_LEN_CHARS_BYTE_SIZE
COMPRESSION_LEVEL_CHUNK_SLICE = HASH_LEN_CHARS_CHUNK_SLICE + COMPRESSION_LEVEL_BYTE_SIZE
COMPRESSION_TYPE_CHUNK_SLICE = COMPRESSION_LEVEL_CHUNK_SLICE + COMPRESSION_TYPE_BYTE_SIZE
EXPRESS_POW_CHUNK_SLICE = COMPRESSION_TYPE_CHUNK_SLICE + EXPRESS_POW_BYTE_SIZE
